'use strict';

/**
 * @typedef {object} Reference
 * @property {string} raw
 * @property {string} action
 * @property {string} prefix
 * @property {string} issue
 * @property {string | null} owner
 * @property {string | null} repository
 */

/**
 * @typedef {object} ParsedCommit
 * @property {string} raw
 * @property {string | null} type
 * @property {string | null} scope
 * @property {string | null} subject
 * @property {string | null} header
 * @property {string | null} body
 * @property {string | null} footer
 * @property {unknown[]} notes
 * @property {Reference[]} references
 * @property {string[]} mentions
 * @property {string | null} merge
 * @property {string | null} revert
 *
 */

/** @typedef {'always' | 'never'} Enable */

const { promises: fs } = require('fs');
const path = require('path');
const { fork } = require('child_process');

/**
 * @param {string} dir
 * @param {string} file
 * @returns {Promise<string>}
 */
const findFileInDirOrAncestor = async (dir, file) => {
  const entries = await fs.readdir(dir);

  if (entries.find(f => f === file)) return path.join(dir, file);

  const newDir = path.join(dir, '..');
  if (newDir === dir) throw new Error(`"${file}" not found.`);

  return findFileInDirOrAncestor(newDir, file);
};

/** @type {string | null} */
let cspellBinPath = null;

module.exports = {
  rules: {
    '@pega/multi-scope-enum': (
      /** @type {ParsedCommit} */ { scope },
      /** @type {Enable} */ enable = 'always',
      /** @type {{ scopes?: string[]; separator?: string | RegExp; trim?: boolean; }} */
      { scopes = [], separator = /\\|\/|,/, trim = false } = {}
    ) => {
      if (!scope || !scopes.length) return [true, ''];

      if (enable === 'always') {
        return [
          scope.split(separator).every(s => scopes.includes(trim ? s.trim() : s)),
          `Scope must be one or more of the following [${scopes.join(', ')}].`
        ];
      }

      // enable === 'never'
      return [
        !scope.split(separator).some(s => scopes.includes(trim ? s.trim() : s)),
        `Scope must not be one or more of the following [${scopes.join(', ')}].`
      ];
    },

    '@pega/header-reference': (
      /** @type {ParsedCommit} */ { header, references },
      /** @type {Enable} */ enable = 'always'
    ) => {
      if (!header) return [true, ''];

      if (enable === 'always') {
        return [
          references.some(({ raw }) => header.includes(raw)),
          'Header must contain a reference id.'
        ];
      }

      // enable === 'never'
      return [
        !references.some(({ raw }) => header.includes(raw)),
        'Header must not contain a reference id.'
      ];
    },

    '@pega/reference-action-enum': (
      /** @type {ParsedCommit} */ { references },
      /** @type {Enable} */ enable = 'always',
      /** @type {{ actions: string[]; caseSensitive: boolean; }} */
      {
        actions = [
          'close',
          'closes',
          'closed',
          'fix',
          'fixes',
          'fixed',
          'resolve',
          'resolves',
          'resolved'
        ],
        caseSensitive = false
      }
    ) => {
      if (!references.length) return [true, ''];

      if (!caseSensitive) actions = actions.map(action => action.toLowerCase());

      if (enable === 'always') {
        return [
          references.some(({ action }) =>
            actions.includes(caseSensitive ? action : action.toLowerCase())
          ),
          `Reference action must be one or more of the following [${actions.join(', ')}].`
        ];
      }

      // enable === 'never'
      return [
        !references.some(({ action }) =>
          actions.includes(caseSensitive ? action : action.toLowerCase())
        ),
        `Reference action must not be one or more of the following [${actions.join(', ')}].`
      ];
    },

    '@pega/reference-order': (
      /** @type {ParsedCommit} */ { references },
      /** @type {Enable} */ enable = 'always',
      { prefixes = ['#'], ascending = true }
    ) => {
      if (references.length < 2 || enable !== 'always') return [true, ''];

      /** @type {{ [prefix: string]: string[][]; }} */
      const referenceMap = {};

      let highestPrefixIndex = -1;
      try {
        references.forEach(({ prefix, issue }) => {
          const idx = prefixes.findIndex(p => p === prefix);

          // Check prefixes are ordered.
          if (highestPrefixIndex > idx) {
            throw new Error(`References must be in the following order [${prefixes.join(', ')}].`);
          }

          // Group reference id's by prefix.
          if (!referenceMap[prefix]) referenceMap[prefix] = [];
          referenceMap[prefix].push(issue.split(/\D+/));

          highestPrefixIndex = idx;
        });
      } catch (e) {
        return [false, e.message];
      }

      // Check id's are ordered.
      // eslint-disable-next-line no-restricted-syntax
      for (const refs of Object.values(referenceMap)) {
        // eslint-disable-next-line no-continue
        if (refs.length < 2) continue;

        if (!ascending) refs.reverse();

        for (let refIdx = 0; refIdx < refs.length - 1; refIdx += 1) {
          const refA = refs[refIdx];
          const refB = refs[refIdx + 1];

          for (let partIdx = 0; partIdx < Math.max(refA.length, refB.length); partIdx += 1) {
            const refAPart = Number.parseInt(refA[partIdx] || '0', 10);
            const refBPart = Number.parseInt(refB[partIdx] || '0', 10);

            if (refAPart < refBPart) break;
            else if (refAPart > refBPart) {
              return [
                false,
                `References must be in ${ascending ? 'ascending' : 'descending'} order.`
              ];
            }
          }
        }
      }

      return [true, ''];
    },

    '@pega/spellcheck': async (
      /** @type {ParsedCommit} */ { raw },
      /** @type {Enable} */ enable = 'always'
    ) => {
      if (enable !== 'always') return [true, ''];

      if (!cspellBinPath) {
        try {
          cspellBinPath = await findFileInDirOrAncestor(
            path.dirname(require.resolve('cspell')),
            'bin.js'
          );
        } catch {
          return [false, 'Unable to load cspell.'];
        }
      }

      const cspell = fork(cspellBinPath, ['--no-summary', '--unique', 'stdin'], {
        stdio: 'pipe'
      });

      cspell.stdin?.write(
        raw
          .split('\n')
          .filter(s => !s.startsWith('#'))
          .join('\n')
      );
      cspell.stdin?.end();

      /** @type {Buffer[]} */
      const out = [];
      let outLen = 0;
      cspell.stdout?.on('data', chunk => {
        out.push(chunk);
        outLen += Buffer.byteLength(chunk);
      });

      return new Promise(resolve => {
        cspell.on('exit', code => {
          if (code === 0) resolve([true, '']);

          const output = Buffer.concat(out, outLen).toString('utf-8');

          console.log(output);

          const words = [...output.matchAll(/^stdin:\d+:\d+ - Unknown word \((.*)\)$/gm)].map(
            ([, word]) => word
          );

          resolve([false, `Unknown word${words.length === 1 ? '' : 's'} [${words.join(', ')}].`]);
        });
      });
    }
  }
};
